"""
Read and validate test file.
"""
import yaml
import json

from os import path
from collections.abc import Sequence

from . import schema 

def find_testfile():
    for filename in ['baygon', 't', 'test', 'tests']:
        for ext in ['json', 'yml', 'yaml']:
            f = f"{filename}.{ext}"
            if path.exists(f):
                return f

def load(filename):
    def loadYaml(filename):
        with open(filename) as fp:
            return yaml.load(fp, Loader=yaml.FullLoader)

    def loadJson(filename):
        with open(filename) as fp:
            return json.load(fp)

    def validate(data):
        return schema.schema(data)

    extension = path.splitext(filename)[1]
    if extension in ['.yml', '.yaml']:
        return validate(loadYaml(filename))
    if extension in ['.json']:
        return validate(loadJson(filename))
    raise ValueError(f'Unknown extension: {extension}')



class Group(Sequence):
    def __init__(self, tests, name=''):
        self.tests = tests
        self.name = name

    def __len__(self):
        return len(self.tests)

    def __repr__(self):
        return '<' + self.__class__.__name__ + ':' + self.name + ':' + repr(self.tests) + '>'

    def __getitem__(self, item):
        return self.tests[item]


class Test(dict):
    def __init__(self, *args, **kwargs):
        super(Test, self).__init__(*args, **kwargs)
        self.__dict__ = self

    @property
    def exit_status(self):
        return self['exit-status'] if 'exit-status' in self else None

    def __repr__(self):
        return '<' + self.__class__.__name__ + ':'  + super().__repr__() + '>'


class Tests(Sequence):
    def __init__(self, filename):
        self.filename = filename
        self.data = load(filename)
        self.name = self.data['name']
        self.filters = self.data['filters'] if 'filters' in self.data else {}
        self.executable = self.data['executable'] if 'executable' in self.data else None
        self.tests = self._build(self.data['tests'])

    def _build(self, tests):
        return [
            Group(self._build(test['tests']), test['name']) if 'tests' in test else Test(test)
            for test in tests
        ]

    def __len__(self):
        return len(self.tests)

    def __repr__(self):
        return '<' + self.__class__.__name__ + ':' + repr(self.tests) + '>'

    def __getitem__(self, item):
        return self.tests[item]



from voluptuous import Schema, ExactSequence, IsFile, Coerce, Required, Any, All, Number, Optional

match = Any(
    All(Number(), Coerce(str), lambda x: [{'equals': x}]),
    All(str, lambda x: [{'equals': x}]),
    [
        All(str, lambda x: {'equals': x}),
        {
            Optional('uppercase'): bool,
            Optional('lowercase'): bool,
            Optional('trim'): bool,
            Required(Any('equals', 'regex', 'contains')): str
        }
    ]
)

test = Schema({
    Optional('name'): str,
    Optional('args'): [Any(str, All(Number(), Coerce(str)))],
    Optional('stdin', default=None): Any(None, str, [str]),
    Optional('stdout', default=[]): match,
    Optional('stderr', default=[]): match,
    Optional('exit-status'): All(Any(int, bool), Coerce(int))
})

group = Schema({
    Optional('name', default=''): str,
    Required('tests'): [test]
})

schema = Schema({
    Required('version'): 1,
    Optional('name', default=''): str,
    Optional('executable'): IsFile('Missing configuration file'),
    Optional('filters'): {
        Optional('uppercase'): bool,
        Optional('lowercase'): bool,
        Optional('trim'): bool,
        Optional('regex'): ExactSequence([str, str])
    },
    Required('tests'): [Any(test, group)]
})



"""
Build a test case for a raw test description
"""
from collections.abc import Sequence

from .executable import Executable
from .validation import Group, Test, Tests

from . import schema, error
import logging

logger = logging.getLogger()


def match(options, value, where=None):
    issues = []

    for case in options:
        if 'uppercase' in case:
            value = value.upper()

        if 'lowercase' in case:
            value = value.lower()

        if 'trim' in case:
            logger.debug('Trimming str')
            value = value.strip()

        if 'regex' in case:
            logger.debug('Checking regex')
            if not value.grep(case['regex']):
                issues += [error.InvalidRegex(value, case['regex'], on=where)]

        if 'contains' in case:
            logger.debug('Checking contains')
            logger.debug(' - value :' + value)
            if case['contains'] not in value:
                logger.debug('New issue : ' + case['contains'])
                issues += [error.InvalidContains(value,
                                           case['contains'], on=where)]
        if 'equals' in case:
            logger.debug('Checking equals')
            if case['equals'] != value:
                issues += [error.InvalidEquals(value, case['equals'], on=where)]

    return issues


class TestCase:
    def __init__(self, options: Test, executable: Executable = None, id=[]):
        self.options = schema.test(options)
        self.name = options.name
        self.executable = executable
        self.exe = self.executable
        self._id = id

    def run(self):

        output = self.exe.run(*self.options.args, stdin=self.options.stdin)
        issues = []
        issues += self._check_exit_status(output)
        issues += self._check_stdout(output)
        issues += self._check_stderr(output)
        return issues

    def _check_exit_status(self, output):
        if self.options.exit_status is None:
            return []
        expected = self.options.exit_status
        if (output.exit_status != expected):
            return [
                error.InvalidExitStatus(
                    output.exit_status, expected
                )
            ]
        return []

    def _check_stdout(self, output):
        if self.options.stdout is None:
            return []
        return match(self.options, output.stdout, where='stdout')

    def _check_stderr(self, output):
        if self.options.stderr is None:
            return []
        return match(self.options, output.stderr, where='stderr')


    @property
    def id(self):
        return '.'.join(map(str, self._id))

    def __repr__(self):
        return f'<TestCase {self.id}:{self.options.name}>'


class TestGroup(Sequence):
    def __init__(self, tests):
        self._tests = tests

    def __len__(self):
        return len(self._tests)

    def __getitem__(self, item):
        return self._tests[item]

    def __repr__(self):
        return repr(self._tests)


class TestSuite(TestGroup):
    def __init__(self, tests: Tests, executable: Executable = None, id=[]):
        self._id = id
        self.name = tests.name
        self._executable = executable if executable is not None else None
        self._tests = list(self._build(tests))

    def _build(self, tests):
        for index, test in enumerate(tests):
            yield [TestCase, TestSuite][isinstance(test, Group)](
                test, self._executable, id=self._id + [index + 1])

    @property
    def id(self):
        return '.'.join(map(str, self._id))

    def run(self):
        return [u.run() for u in self._tests]



class InvalidCondition:
    def __init__(self, value, expected, message=None, on=None):
        self.value = value
        self.expected = expected
        self.message = message
        self.on = on

    def __str__(self):
        return f'Expected "{self.expected}", but got "{self.value}"'


class InvalidExitStatus(InvalidCondition):
    def __str__(self):
        if hasattr(self.value, '__len__') and len(self.value) > 20:
            return f'Invalid exit status. Expected {self.expected}'
        else:
            return f'Invalid exit status. Expected {self.expected}, but got {self.value}'


class InvalidContains(InvalidCondition):
    pass


class InvalidRegex(InvalidCondition):
    def __str__(self):
        return f'Invalid value on {self.on}. Expected to match regex /{self.expected}/'


class InvalidEquals(InvalidCondition):
    def __str__(self):
        if hasattr(self.value, '__len__') and len(self.value) > 20:
            return f'Invalid value on {self.on}. Expected exactly "{self.expected}"'
        else:
            return f'Invalid value on {self.on}. Expected exactly "{self.expected}", but got "{self.value}"'




import os
import re
import subprocess

from collections import namedtuple


Outputs = namedtuple('Outputs', ['exit_status', 'stdout', 'stderr'])


class Executable:
    """ Allow to execute a program and conveniently read the output. """
    def __init__(self, filename, encoding='utf-8', filters={}):
        self.filename = filename
        self.encoding = encoding
        self.filters = filters

        if not self._is_executable(filename):
            raise ValueError("Program %s is not executable!" % filename)

    def run(self, *args, stdin=None):
        p = subprocess.Popen([self.filename, *[str(a) for a in args]],
                             stdout=subprocess.PIPE,
                             stdin=subprocess.PIPE,
                             stderr=subprocess.PIPE)

        stdout, stderr = p.communicate(input=stdin)

        stdout = stdout.decode(self.encoding) if stdout is not None else None
        stderr = stderr.decode(self.encoding) if stderr is not None else None

        stdout = self._filter(self.filters, stdout)
        stderr = self._filter(self.filters, stderr)

        return Outputs(p.returncode, GreppableString(stdout), GreppableString(stderr))

    def __call__(self, *args, **kwargs):
        return self.run(*args, **kwargs)

    @staticmethod
    def _is_executable(filename):
        return os.path.isfile(filename) and os.access(filename, os.X_OK)

    def _filter(self, filter, value):
        if 'uppercase' in filter:
            value = value.upper()

        if 'lowercase' in filter:
            value = value.lower()

        if 'trim' in filter:
            value = value.strip()

        if 'regex' in filter:
            value = re.sub(filter['regex'][0], filter['regex'][1], value)

        return value


class GreppableString(str):
    """ A string that can be parsed with regular expressions. """
    def grep(self, pattern):
        return re.findall(pattern, self)



import click
import time
import json
import logging
import os


from . import TestSuite, TestCase, Tests, validation, Executable


def pad_size(tests):
    width, _ = click.get_terminal_size()
    max_test_len = 0
    for t in tests:
        if len(t.name) > max_test_len:
            max_test_len = len(t.name)
    max_display_len = max_test_len + len('Test %d: PASSED' % len(tests))
    if width > max_display_len:
        return max_display_len
    return 0


def display_pad(pad=0):
    if pad == 0:
        click.secho('  ', nl=False)
    else:
        click.secho('.' * pad, nl=False, dim=True)


def display_pass(test, pad=0):
    display_pad(pad - len(test.name) - len(str(test.id)))
    click.secho(' PASSED', fg='green')


def display_fail(test, pad=0):
    display_test_name(test)
    display_pad(pad - len(test.name) - len(str(test.id)))
    click.secho(' FAILED', fg='red')


def test_name_length(test):
    pad = '  ' * (len(test.id) - 1)
    return len(f'{pad}Test {test.id}: {test.name}')


def display_test_name(test):
    pad = '  ' * (len(test.id) - 1)
    click.secho(f'{pad}Test {test.id}: ', nl=False, bold=True)
    click.echo(f'{test.name}', nl=False)


def display_failure(test, failures, pad, verbose=0):
    display_fail(test, pad)
    for f in failures:
        click.secho('  - ' + str(f), fg='yellow')
        if verbose > 1:
            click.secho(f.got, fg='cyan')


class OneLineExceptionFormatter(logging.Formatter):
    def formatException(self, exc_info):
        result = super().formatException(exc_info)
        return repr(result)

    def format(self, record):
        result = super().format(record)
        if record.exc_text:
            result = result.replace("\n", "")
        return result


class Runner:
    def __init__(self, verbose, executable=None, **kwargs):
        if verbose > 3:
            self._init_logger("DEBUG")

        self.verbose = verbose
        self.executable = executable
        self.limit = -1 if 'limit' not in kwargs else kwargs['limit']
        self.filename = validation.find_testfile()
        self.test_description = Tests(self.filename)
        self.test_suite = TestSuite(
            self.test_description, Executable(self.executable))

    def _init_logger(self, loglevel):
        handler = logging.StreamHandler()
        formatter = OneLineExceptionFormatter(logging.BASIC_FORMAT)
        handler.setFormatter(formatter)
        root = logging.getLogger()
        root.setLevel(os.environ.get("LOGLEVEL", loglevel))
        root.addHandler(handler)

    def run(self):
        start_time = time.time()
        self._traverse_group(self.test_suite)

        click.secho('\nRan %d tests in %ss.' % (
            self.successes + self.failures,
            round(time.time() - start_time, 2)
        ), bold=True)
        
        if self.failures > 0:
            click.echo('%d failed, %d passed (%d%% ok).' % (
                self.successes + self.failures, self.failures),
                100-round(self.failures/self.successes*100, 2))
            click.secho('\nfail.', fg='red')
        else:
            click.secho('\nok.', fg='green')        

        return self.failures
        
    def _max_length(self, tests):
        length = 0
        for test in tests:
            if isinstance(test, TestSuite):
                length = max(test_name_length(test), length)
                length = max(self._max_length(test), length)
            else:
                length = max(test_name_length(test), length)
        return length

    def _traverse_group(self, tests):
        pad = self._max_length(tests) + 10

        self.failures = 0
        self.successes = 0

        for test in tests:
            if self.limit > 0 and self.failures > self.limit: 
                break
        
            if isinstance(test, TestSuite):
                display_test_name(test)
                click.echo('')
                self._traverse_group(test)
            elif isinstance(test, TestCase):
                display_test_name(test)
                issues = test.run()
                display_pad(pad - test_name_length(test))
                if not len(issues):
                    self.successes += 1
                    click.secho(' PASSED', fg='green')
                else:
                    self.failures += 1
                    click.secho(' FAILED', fg='red')
                    for issue in issues:
                        click.secho(repr(issue.value), fg='cyan')
                        click.secho('  ' * len(test.id) +
                                    str(issue), fg='yellow')
        return self.failures


@click.command()
@click.argument('executable', required=False, type=click.Path(exists=True))
@click.option('-v', '--verbose', count=True)
@click.option('-l', '--limit', type=int, default=-1)
def cli(verbose=0, executable=None, **kwargs):
    runner = Runner(verbose, executable, **kwargs)
    failures = runner.run()
    click.echo('')
    return failures


if __name__ == '__main__':
    cli()



__version__ = '0.1.1'

__author__ = 'Yves Chevallier'
__description__ = 'Functional tests for teaching activity'
__email__ = 'yves.chevallier@heig-vd.ch'
__license__ = 'MIT'
__copyright__ = 'Copyright 2020 HEIG-VD'

from .validation import Test, Tests
from .test import TestCase, TestSuite
from .executable import Executable



import setuptools
import re

with open("README.md", "r") as fh:
    long_description = fh.read()

name='baygon'

def get_property(prop, project):
    result = re.search(r'__{}__\s*=\s*[\'"]([^\'"]*)[\'"]'.format(prop), open(project + '/__init__.py').read())
    return result.group(1)

setuptools.setup(
    name=name,
    version=get_property('version', name),
    author=get_property('author', name),
    author_email=get_property('email', name),
    description=get_property('description', name),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/heig-tin-info/baygon",
    install_requires=["voluptuous", "click", "pyaml", "colorama"],
    packages=setuptools.find_packages(),
    license="MIT",
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    py_modules=[name],
    entry_points={"console_scripts": [f"{name}={name}.__main__:cli"]},
)
